[mypy]
warn_unused_ignores = True
check_untyped_defs = True
disable_error_code = import-untyped, import-not-found, unused-ignore

[mypy-stringcase.*]
ignore_missing_imports = True

[mypy-jacobsjinjatoo.*]
ignore_missing_imports = True

[mypy-yamlloader.*]
ignore_missing_imports = True

[mypy-textual.*]
ignore_missing_imports = True

[mypy-pyqttier.*]
ignore_missing_imports = True

[mypy-isodate.*]
ignore_missing_imports = True

[mypy-stinger_python_utils.*]
ignore_missing_imports = True